;----------------------------------------------------------------------------;
; PIC-Controller software for HRP137 Receiver
;
; This RX was designed by Jaap van Rotten and Arne van Belle.
; A PIC controller allows for single knop operation of HPR137 RX and Alblas
; decoder or alternately remote control via RS232. So you can build a fully automatic
; station.
; Up to 7 channels my be switched via hardware rotary switch
; Up to 8 channels my be switched via RS232 remote control
; most recent setting either by hardware or by remote control will override previous setting.
; There are two outputs from PIC to directly control Alblas decoder, with some
; additional glue-electronics you may also switch the relay inside HPR137.
; Syntax for remote control is: Chn with n=0 to 7 ( Ch1,Ch2 etc. ). This allows for sharing
; the same serial port with the rotor controller using Easycomm or GS232 protocol.
;
; Sofware is based on a piece of I2C code from an anonymous author.
; This software has been structured somewhat and adapted as follows
;  
; 1. I2C Controller for UV816/UV916 type tuner with synth IC TSA5510.  
; 2. put switches to work compatible with Elsdon RX, 7 channels available
; 3. change tuner frequencies: my converter translates 1691 MHz to 137.5 MHz
; 4. add control of Alblas decoder                      
; 5. add serial input for remote control, 8 channels available, 19200,8,1
;
;
;	Copyright by Goetz Romahn November 2002
;----------------------------------------------------------------------------;
        PROCESSOR 16f84
        #include "p16f84.inc"
        #include "bank.inc"	; Assistant files
	LIST
	__CONFIG _CP_OFF & _PWRTE_ON & _XT_OSC & _WDT_OFF

;variables start at 0x0C
	Cblock 0x0C
;
; interrupt context save 
	w_
	status_
	fsr_
	pclath_
;
; synthesizer and switch interface
	New_Switch
	last_Switch
	loop_count
	Channel_No
	acc_lo
	acc_hi
	i2c_data
;
; wait macro
	WCYCLE
	PRESCwait
;
; rs232 register
	RS_TEMP1	; Temp register
	RS_TEMP2
	RXD
	fifo_char
	fifo_len
	fifo_start
	fifo_base
	fifo_buffer:16
        endc
;
; definitions
I2C_Add		equ	.194		;Tuner I2C Address (common address used)
I2C_Control	equ	.142		;Sent to synth to set up functions etc
I2C_Band	equ	.96		;Band switch settings
Offset_Lo_Byte 	equ    	.240  		;175 MHz offset; offset Hi+Lo = .2800
Offset_Hi_Byte 	equ    	.10     	;2800*0.0625=175 MHz               
scl           	equ    	0		;Port A bit 0
sda           	equ    	1		;Port A bit 1
scl_out_sda_out equ 	b'00001000'
scl_out_sda_in  equ    	b'00001010'
scl_in_sda_in   equ    	b'00001011'
;-------------------------------------;
;PIN ASSIGNMENTS

;Pin Number	Port	Use
;	1	RA2	output:switch Alblas decoder, HRPT to PDUS/C-HRPT 
;	2	RA3	not used ( defect on my PIC )
;	3	RA4	switch Alblas decoder, PDUS to C-HRPT while RA2=1
;	4	RESET	Tie to +Ve
;	5	Vss	Ground
;	6	RB0	Serial in ( inverted )	
;	7	RB1	Switch1 in	
;	8	RB2	Switch2 in
;	9	RB3	Switch3 in
;	10	RB4	Switch4 in
;	11	RB5	Switch5 in
;	12	RB6	Switch6 in
;	13	RB7	not used
;	14	Vdd	+Ve supply
;	15	Osc	Xtal 4 MHz
;	16	Osc	Xtal
;	17	RA0	SCL	Pull up with 10K
;	18	RA1	SDA	Pull up with 10K
;	Switch common goes to ground
;	no pull ups on switch required
;-----------------------------------------------------------------------------
	org	0x00            ; start of main program
	goto Init
	org	0x04		;interrupt vector
	goto	ISR
;-----------------------------------------------------------------------------
; put jump tables to a safe place, do not move them unless you know the consequences
;-----------------------------------------------------------------------------
Switch_table		; get switch #
	addwf  	PCL,f          	;valid Switch inputs	
	retlw  	b'11111111'   	;Switch 0	; impossible value, HRPT 1707
 	retlw  	b'01111100'	;Switch 1	; HRPT		1698
 	retlw  	b'01111010'	;Switch 2	; C-HRPT	1700.4
 	retlw  	b'01110110'	;Switch 3	; HRPT		1702.5
	retlw  	b'01101110'   	;Switch 4	; PDUS2		1694.5
 	retlw  	b'01011110'	;Switch 5	; PDUS1		1691
 	retlw  	b'00111110'	;Switch 6	; HRPT		1707
 	retlw  	b'01111110'	;Switch 7	; HRPT		1707 all contacts open
;-------------------------------------;
channel_table		; frequency setting
			; my converter translates 1691MHz to 137.5 MHz
			; 2800 == 175 MHz offset
	addwf  	PCL,f	;channel no table
	retlw  	.243	;Fequency = (243+2800) * 0.0625 MHz ;freq 1707   IF 36.7; tuned to 153.5 MHz
	retlw  	.99	;Fequency = (099+2800) * 0.0625 MHz ;freq 1698	 IF 36.7; tuned to 144.5 MHz 
	retlw  	.138	;Fequency = (138+2800) * 0.0625 MHz ;freq 1700.4 IF 36.7; tuned to 146.9 MHz 
	retlw  	.171	;Fequency = (171+2800) * 0.0625 MHz ;freq 1702.5 IF 36.7; tuned to 149.0 MHz
	retlw  	.78	;Fequency = (078+2800) * 0.0625 MHz ;freq 1694.5 IF 38.9; tuned to 141.0 MHz 
	retlw  	.22	;Fequency = (022+2800) * 0.0625 MHz ;freq 1691   IF 38.9; tuned to 137.5 MHz
	retlw  	.243	;Fequency = (243+2800) * 0.0625 MHz ;freq 1707   IF 36.7; tuned to 153.5 MHz 
	retlw  	.243	;Fequency = (243+2800) * 0.0625 MHz ;freq 1707   IF 36.7; tuned to 153.5 MHz 
;----------------------------------------------------------------------------;
decoder_table		; PORTA settings for Alblas decoder
	addwf  	PCL,f	;channel no table
	retlw  	b'00000000'	;HRPT
	retlw  	b'00000000'	;HRPT
	retlw  	b'00010100'	;C-HRPT
	retlw  	b'00000000'	;HRPT
	retlw  	b'00000100'	;PDUS
	retlw  	b'00000100'	;PDUS
	retlw  	b'00000000'	;HRPT
	retlw  	b'00000000'	;HRPT
;************************************************************************

        #INCLUDE WAIT.inc
        #include "rs232.inc"
;************************************************************************
;       Interrupt service fuer RS232

ISR     movwf   w_		; save complete context
        swapf   STATUS, w
        movwf   status_
        bcf     STATUS, RP0
	movfw	PCLATH
	movwf	pclath_
        movfw   FSR
	movwf	fsr_

        RECEIVE

ISRend
	movfw	fsr_		; restore context
        movwf   FSR
	movfw	pclath_
	movwf	PCLATH
        swapf   status_,w
        movwf   STATUS
        swapf   w_,f
        swapf   w_,w
        bcf     INTCON,INTF
        RETFIE

;************************************************************************
; start of program

Init
        WAIT    0x80            ;wait for Alblas Eprom to load
	movlw	scl_in_sda_in	;set scl and sda as input
	BANK1
	movwf	TRISA
	BANK0
	movwf  	PORTA		;default Alblas decoder to HRPT 
        WAIT    0x80            ;wait for Alblas Eprom to load
	clrf	last_Switch
        RS232init       ; RS232

;************************************************************************
; endless loop is here
Get_Switch
	movfw  	PORTB		;get the Switch input
	andlw	0x7e		;mask out bits 0 and 7
	movwf  	New_Switch	;Store port values
	xorwf  	last_Switch,w	;check to see if its 
	btfss  	STATUS,Z	;the same as last time
	goto   	Switch_Diff
;
	movfw   fifo_len	;now check for serial input via interrupt
        btfsc   STATUS,Z
	goto   Get_Switch	; no switch, no serial input: idle

;************************************************************************
;rs232 service			;legal strings are "Chx" where x=0,1,2,3,4,5,6,7

	call    FIFOget
	movlw   'C'		;string must start with 'C'
        subwf   fifo_char,w
        bz      RS_h
	goto   Get_Switch

RS_h
;to avoid system hang when serial data stops prematurely we continuously poll for new switch		
	movfw  	PORTB		;get the Switch input
	andlw	0x7e		;mask out bits 0 and 7, bit 0 serial port, bit 7 unused
	movwf  	New_Switch	;Store port values
	xorwf  	last_Switch,w	;check to see if its 
	btfss  	STATUS,Z	;the same as last time
	goto   	Switch_Diff
;

	movfw   fifo_len	;now check for serial input via interrupt
        btfsc   STATUS,Z
	goto	RS_h		;wait for next character
        call    FIFOget
        movlw   'h'		;next character must be 'h'
        subwf   fifo_char,w
        bz      RS_channel	;no further syntax check, we just jump
	goto   Get_Switch

;select channel via RS232
RS_channel
;to avoid system hang when serial data stops prematurely we continuously poll for new switch		
	movfw  	PORTB		;get the Switch input
	andlw	0x7e		;mask out bits 0 and 7, bit 0 serial port, bit 7 unused
	movwf  	New_Switch	;Store port values
	xorwf  	last_Switch,w	;check to see if its 
	btfss  	STATUS,Z	;the same as last time
	goto   	Switch_Diff
;
	movfw   fifo_len	;now check for serial input via interrupt
        btfsc   STATUS,Z
	goto	RS_channel
        call    FIFOget
	andlw	0x07		; there is no check for legal characters, 
	movwf	loop_count	; we just force legal values
	goto	Switch_found

;----------------------------------------------------------------------------;
;switch service

Switch_Diff
	movfw 	New_Switch
	movwf  	last_Switch	;Save Switch input for future reference
	movlw  	.8		;Check 8 switch possitions
	movwf  	loop_count
Find_Switch
	movfw	loop_count  	;look for Switch in table
	call   	Switch_table
	xorwf  	last_Switch,w   ;Switch same as looked up?
	btfsc  	STATUS,Z
	goto   	Switch_found    ;yes
	decfsz 	loop_count,f  	;no go round again! not expected to reach zero
	goto   	Find_Switch	;
;----------------------------------------------------------------------------;
; common code for switch and RS232

Switch_found
	movfw	loop_count  	;find channel no
	call   	decoder_table
	movwf	PORTA		;switch Alblas decoder accordingly
        WAIT    0x80            ;wait for Alblas Eprom to load
	movfw	loop_count  	;find channel no
	call   	channel_table
	movwf  	Channel_No
	goto   	Rc_Routine	;just for readability

;----------------------------------------------------------------------------;
;set synthesizer via I2C

Rc_Routine
	movlw	Offset_Lo_Byte	;Add offset to channel no
	movwf	acc_lo
	movlw	Offset_Hi_Byte
	movwf	acc_hi
	movfw  	Channel_No  
	addwf  	acc_lo,f      	
	btfsc  	STATUS,C	;test carry 
	incf   	acc_hi,f      	;if set increment acc_hi
	call   	I2C_Start
	movlw  	I2C_Add   	;synth address
	call   	I2C_Tx
	movfw   acc_hi       	;synth hi byte
	call   	I2C_Tx
	movfw   acc_lo       	;synth low byte 
	call   	I2C_Tx
	movlw  	I2C_Control   	;synth command
	call   	I2C_Tx
	movlw  	I2C_Band	;synth band select
	call   	I2C_Tx
	call   	I2C_Stop
	goto   Get_Switch	;and continue looping

;----------------------------------------------------------------------------;
; I2C code, it works why should I change it

I2C_Start
	movlw  	scl_out_sda_out      	;set scl and sda as output
	BANK1
	movwf	TRISA
	BANK0	
	bsf    	PORTA,sda     
	nop
	bsf    	PORTA,scl
	nop
	bcf    	PORTA,sda     ;set start condition
	nop
	bcf    	PORTA,scl
	nop
	retlw  	00
I2C_Tx
	movwf	i2c_data
	movlw  	scl_out_sda_out      	;set scl and sda as output
	BANK1
	movwf	TRISA
	BANK0
	movlw  	.8            		;loop 8 times (8 bits)
	movwf  	loop_count
I2C_Tx_Loop
	rlf    	i2c_data,f    		;rotate into carry
	bsf    	PORTA,sda
	btfss  	STATUS,C
	bcf    	PORTA,sda
	nop
	bsf    	PORTA,scl     		;clock data bit 
	nop
	bcf    	PORTA,scl     		;out
	nop
	decfsz 	loop_count,1
	goto   	I2C_Tx_Loop
	movlw  	scl_out_sda_in       	;set sda as input
	BANK1
 	movwf	TRISA
	BANK0
	bsf    	PORTA,scl     		;aknowledge
	nop
	bcf    	PORTA,scl
	nop
	retlw  	00
I2C_Stop
	movlw  	scl_out_sda_out      	;set scl and sda as output
	BANK1
	movwf	TRISA
	BANK0
	bcf    	PORTA,sda
	nop
	bsf    	PORTA,scl
	nop
	bsf    	PORTA,sda
	nop
	movlw  	scl_in_sda_in        	;set scl and sda as input
	BANK1
	movwf	TRISA
	BANK0
	retlw  	00

	end
